package 
{
	import com.mousman.as3.agal.light.IMLight;
	import com.mousman.as3.agal.mesh.IMMesh;
	import com.mousman.as3.agal.parser.dto.MBuffersWriterDTO;
	import com.mousman.as3.agal.parser.objs.MMtlSet;
	import com.mousman.as3.agal.parser.objs.MObject3D;
	import com.mousman.as3.agal.parser.objs.MObjMesh;
	import com.mousman.as3.agal.parser.objs.MObjParser;
	import com.mousman.as3.agal.parser.objs.MPolygoneGroup3D;
	import com.mousman.as3.agal.parser.objs.MPolygonGroupBuffers;
	import com.mousman.as3.agal.parser.shaders.ShaderSet;
	import com.mousman.as3.agal.parser.shaders.ShaderStore;
	import com.mousman.as3.agal.parser.verticeswriter.MVerticesBytesArrayWriter;
	import com.mousman.as3.agal.shader.MBasicShader;
	import com.mousman.as3.agal.shader.MPhongShader;
	import com.mousman.as3.agal.shader.MSimpleShader;
	import com.mousman.as3.agal.shader.ShaderType;
	import com.mousman.as3.utils.color.ColorUtil;
	import flash.display3D.Context3D;
	import flash.display3D.Context3DBlendFactor;
	import flash.display3D.Context3DProgramType;
	import flash.display3D.IndexBuffer3D;
	import flash.display3D.Program3D;
	import flash.display3D.VertexBuffer3D;
	import flash.geom.Matrix3D;
	import flash.geom.Vector3D;
	import flash.utils.ByteArray;
	
	/**
	 * ...
	 * @author Moussa Dembélé (www.mousman.com)
	 */
	public class CupsMesh implements IMMesh
	{
		[Embed(source="../lib/cups/cups.obj", mimeType="application/octet-stream")]
		private var _objData:Class;

		[Embed(source="../lib/cups/cups.mtl", mimeType="application/octet-stream")]
		private var _mtlData:Class;	
		
		// contains all the ShaderNCo
		private var _shaderStore:ShaderStore;
		
		private var _context3D:Context3D;
		
		private var _parser:MObjParser;
		private var _mesh:MObjMesh;
		private var _modelMatrix:Matrix3D;
		private var _modelPosition:Vector3D;
		private var _modelRotation:Vector3D;
		
		private var _modelViewProjection:Matrix3D;
		
		private var _lights:Vector.<IMLight>;
		private var _cameraPosition:Vector3D;
		private var _cameraMatrix:Matrix3D;
		private var _ambiantLight:Vector.<Number>;
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		public function CupsMesh(context3D:Context3D){
			
			_context3D = context3D;
			_modelMatrix = new Matrix3D();
			_modelPosition = new Vector3D();
			_modelRotation = new Vector3D();
			_lights = new Vector.<IMLight>();

			// definition and decalration of the shaders to use
			setShaders();
			
			// parse the OBJ file and create buffers
			_parser = new MObjParser(_context3D, 5, _shaderStore);
			var bytes:ByteArray = new _objData();
			var objFile:String = bytes.readUTFBytes(bytes.bytesAvailable);			
			bytes = new _mtlData();
			var mtlFile:String = bytes.readUTFBytes(bytes.bytesAvailable);
			_mesh = _parser.parse(objFile, mtlFile);
			
			// assemble and upload program
			initPrograms();
			
			count();
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------
		private function count():void{
			var object3Dset:MObject3D;
			var mPolygonsGroup:MPolygoneGroup3D;
			var polygonBufferSet :MPolygonGroupBuffers;
			var totalVertices:int;
			var totalObjectVertices:int ;
			
			for each(object3Dset in _mesh.mObjects) {
				//trace("--------  object " + object3Dset.id);
				totalObjectVertices = 0;
				for each(mPolygonsGroup in object3Dset.polygonsGroups) {
					//trace("   polygon group : " + mPolygonsGroup.id + "    nbVertices : " +mPolygonsGroup.nbVertices);
					//trace("      use texture : " + mPolygonsGroup.mtlSet.ambiantTexture + " , " + mPolygonsGroup.mtlSet.diffuseTexture);
					totalObjectVertices += mPolygonsGroup.nbVertices;
					for each(polygonBufferSet in mPolygonsGroup.polygonBufferSet) {
						//trace("MPolygonGroupBuffers");
					}
					
				}
				totalVertices += totalObjectVertices;
				trace("total vertices for object " + object3Dset.id +" : " + totalObjectVertices);
			}
			trace("total vertices : " + totalVertices);
			
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		private function setShaders():void {
			// ShaderSets's container
			_shaderStore = new ShaderStore();
			
			// definition and declaration of the shaders to use
			var shader:MPhongShader = new MPhongShader();
			var writer:MVerticesBytesArrayWriter = new MVerticesBytesArrayWriter();
			// a ShaderSet object contains a IMShader and a IMBytesArrayWriter
			var shaderSet:ShaderSet = new ShaderSet(ShaderType.DEFAULT_SHADER, shader, writer);
			_shaderStore.add(shaderSet);
			
			shaderSet = new ShaderSet("simpleShader", new MSimpleShader(), writer);
			_shaderStore.add(shaderSet);			
			
			shaderSet = new ShaderSet("basicShader", new MBasicShader(), writer);
			_shaderStore.add(shaderSet);		
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		private function initPrograms():void {
			// create and upload programs
			for each (var shaderSet:ShaderSet in _shaderStore.shaderSetArray) {
				shaderSet.assembleNupload(_context3D);
			}
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		public function addLight(light:IMLight):void {
			_lights.push(light);
		}		
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		public function setSaticConstants():void {
			var color:Object;
			var pos:Vector.<Number>;
			//fc0, ambient light 
			_context3D.setProgramConstantsFromVector(Context3DProgramType.FRAGMENT, 0, _ambiantLight);
			//fc1, light diffuse color
			color = ColorUtil.hexArgb(_lights[0].diffuseColor);
			_context3D.setProgramConstantsFromVector(Context3DProgramType.FRAGMENT, 1, Vector.<Number>([Number(color.rouge) / 255, Number(color.vert) / 255, Number(color.bleu) / 255, _lights[0].diffuseIntensity]));
			//fc2, light position	
			pos =  Vector.<Number>([_lights[0].position.x,_lights[0].position.y,_lights[0].position.z,1]);
			_context3D.setProgramConstantsFromVector(Context3DProgramType.FRAGMENT, 2, pos);
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		private function setFrameConstants():void {
			//vc0, scene
			_context3D.setProgramConstantsFromMatrix(Context3DProgramType.VERTEX, 0, _modelViewProjection, true);
			//vc4, object position
			_context3D.setProgramConstantsFromMatrix(Context3DProgramType.VERTEX, 4, _modelMatrix, true); 
			//vc8, object rotation
			var rotMat:Matrix3D = _modelMatrix.clone();
			rotMat.transpose();
			rotMat.invert();			
			_context3D.setProgramConstantsFromMatrix(Context3DProgramType.VERTEX, 8, rotMat, true);	
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		public function drawTriangles():void {
			var shaderSet:ShaderSet;
			var previousShaderSet:ShaderSet;
			var buffer3D:VertexBuffer3D;
			var index3D:IndexBuffer3D;
			var object3Dset:MObject3D;
			var mPolygonsGroup:MPolygoneGroup3D;
			var polygonBufferSet:MPolygonGroupBuffers;
			
			// define the constantes wich are not dependant of objects or groups
			setFrameConstants();
			
			// loop through the objects
			var oSize:int = _mesh.mObjects.length;
			for (var i:int = 0; i < oSize; i++ ) {
				object3Dset =  _mesh.mObjects[i];
				//trace("drawing object " + object3Dset.id + " , " +object3Dset.id.length);
				// define the programs depending on groups
				previousShaderSet = shaderSet;
					switch(object3Dset.id) {
						case "sugar1" :
						case "sugar2" :
							shaderSet = _shaderStore.getShaderById("basicShader");
							break;
						case "sugar3" :
						case "sugar4" :
						case "chocolate" :
						case "cofee" :
							shaderSet = _shaderStore.getShaderById("simpleShader");
						break;
						default :
						    shaderSet = _shaderStore.getShaderById(ShaderType.DEFAULT_SHADER);
						break;
					}
				if (previousShaderSet != shaderSet) {
					if(previousShaderSet) 
						previousShaderSet.shader.releaseActiveVertextBuffer(_context3D);
					_context3D.setProgram(shaderSet.program);	
				}
				// loop through the groups
				var pgSize:int = object3Dset.polygonsGroups.length;
				for (var j:int = 0; j < pgSize; j++ ) {
					mPolygonsGroup = object3Dset.polygonsGroups[j];
					//trace("drawing group " + mPolygonsGroup.id);
						
					// define the constantes depending on groups
					switch(object3Dset.id) {
						case "sugar1" :
						case "sugar2" :
						case "sugar3" :
						case "sugar4" :
						case "chocolate" :
						case "cofee" :							
							if(previousShaderSet != shaderSet)
								setSimpleConstants(mPolygonsGroup.mtlSet);
						break;
						default :
							if(previousShaderSet != shaderSet)
								setDefaultConstants(mPolygonsGroup.mtlSet);
						break;
					}
					
					// loop through the buffers
					var pbSize:int = mPolygonsGroup.polygonBufferSet.length;
					for (var k:int = 0; k < pbSize; k++ ) {
						polygonBufferSet = mPolygonsGroup.polygonBufferSet[k];
						buffer3D = polygonBufferSet.vertexBuffer;
						index3D = polygonBufferSet.indexBuffer;
						shaderSet.shader.setActiveVertextBuffer(_context3D, buffer3D);
						//trace("drawing triangles ");
						_context3D.drawTriangles(index3D, 0, -1);
					}
				}
			}
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		public function setDefaultConstants(mtlSet:MMtlSet):void {
			var pos:Vector.<Number>;
			
			//fc3, light specular color
			color = ColorUtil.hexArgb(_lights[0].specularColor);
			_context3D.setProgramConstantsFromVector(Context3DProgramType.FRAGMENT, 3, Vector.<Number>([Number(color.rouge)/255,Number(color.vert)/255,Number(color.bleu)/255,_lights[0].specularIntensity]));
			//fc4, camera position	
			pos =  Vector.<Number>([_cameraPosition.x, _cameraPosition.y, _cameraPosition.z, 1]);
			_context3D.setProgramConstantsFromVector(Context3DProgramType.FRAGMENT, 4, pos);			
			// fc5 , material ambiant color
			var color:Object = ColorUtil.hexArgb(mtlSet.ambiantColor);
			_context3D.setProgramConstantsFromVector(Context3DProgramType.FRAGMENT,  5, Vector.<Number>([Number(color.rouge)/255,Number(color.vert)/255,Number(color.bleu)/255,1]));
			//fc6, specular material color
			color = ColorUtil.hexArgb(mtlSet.specularColor);
			_context3D.setProgramConstantsFromVector(Context3DProgramType.FRAGMENT, 6, Vector.<Number>([Number(color.rouge)/255,Number(color.vert)/255,Number(color.bleu)/255,mtlSet.specularCoef]));
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		public function setSimpleConstants(mtlSet:MMtlSet):void {
			// fc3 , material ambiant color
			var color:Object = ColorUtil.hexArgb(mtlSet.ambiantColor);
			_context3D.setProgramConstantsFromVector(Context3DProgramType.FRAGMENT,  6, Vector.<Number>([Number(color.rouge) / 255, Number(color.vert) / 255, Number(color.bleu) / 255, 1]));
		}		
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		public function setProgramConstantsFromMatrix(type:String,firstRegister:int, matrix:Matrix3D, transposedMatrix:Boolean=false):void {
			_context3D.setProgramConstantsFromMatrix(type, firstRegister, matrix, transposedMatrix);
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		public function setProgramConstantsFromVector(type:String,firstRegister:int, data:Vector.<Number>, numRegisters:int=-1):void {
			_context3D.setProgramConstantsFromVector(type, firstRegister, data, numRegisters);
		}	
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		public function setVertexBufferAt(index:int, vertexBuffer:VertexBuffer3D, bufferOffset:int, format:String):void {
			_context3D.setVertexBufferAt(index, vertexBuffer, bufferOffset, format);
		}		
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		public function destroy():void {
			_shaderStore.destroy();
			_shaderStore = null;
			_parser.destroy();
			_parser = null;
			_mesh.destroy();
			_mesh = null;
			_lights.length = 0;
			_lights = null;
			
			context3D.setVertexBufferAt(0, null);
			context3D.setVertexBufferAt(1, null);
			context3D.setVertexBufferAt(2, null);
			context3D.setVertexBufferAt(3, null);			
			context3D.setVertexBufferAt(4, null);			
			context3D.setVertexBufferAt(5, null);			
			context3D.setVertexBufferAt(6, null);			
			context3D.setVertexBufferAt(7, null);			
		}			
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------	
		public function get position():Vector3D {return _modelPosition;	}
		public function set position(value:Vector3D):void {
			_modelPosition.x = value.x;
			_modelPosition.y = value.y;
			_modelPosition.z = value.z;
			
			_modelMatrix.identity();
			
			_modelMatrix.appendRotation(_modelRotation.x, Vector3D.X_AXIS);
			_modelMatrix.appendRotation(_modelRotation.y, Vector3D.Y_AXIS);
			_modelMatrix.appendRotation(_modelRotation.z, Vector3D.Z_AXIS);
			
			_modelMatrix.appendTranslation(_modelPosition.x, _modelPosition.y, _modelPosition.z);
		}
		public function get rotation():Vector3D {return _modelRotation;}
		public function setRotation(rx:Number, ry:Number, rz:Number):void {
			_modelMatrix.identity();
			_modelRotation.x = rx;
			_modelRotation.y = ry;
			_modelRotation.z = rz;
			
			_modelMatrix.appendRotation(_modelRotation.x, Vector3D.X_AXIS);
			_modelMatrix.appendRotation(_modelRotation.y, Vector3D.Y_AXIS);
			_modelMatrix.appendRotation(_modelRotation.z, Vector3D.Z_AXIS);
			
			_modelMatrix.appendTranslation(_modelPosition.x, _modelPosition.y, _modelPosition.z);
		}
		
		public function get modelMatrix():Matrix3D {return _modelMatrix;}
		public function set modelMatrix(value:Matrix3D):void {_modelMatrix = value;}

		public function get context3D():Context3D {return _context3D;}
		public function set context3D(value:Context3D):void {_context3D = _context3D;}	
		
		public function get modelViewProjection():Matrix3D {return _modelViewProjection;}
		public function set modelViewProjection(value:Matrix3D):void {_modelViewProjection = value;}
		
		public function get cameraPosition():Vector3D {	return _cameraPosition;}
		public function set cameraPosition(value:Vector3D):void {_cameraPosition = value;}
		
		public function get ambiantLight():Vector.<Number> {return _ambiantLight;}
		public function set ambiantLight(value:Vector.<Number>):void {_ambiantLight = value;}
		
		public function set cameraMatrix(value:Matrix3D):void {	_cameraMatrix = value;	}
		
	}
	
}