package  
{
	import com.adobe.utils.PerspectiveMatrix3D;
	import com.mousman.as3.agal.camera.IMCamera;
	import com.mousman.as3.agal.camera.OrbitalCamera;
	import com.mousman.as3.agal.light.IMLight;
	import com.mousman.as3.agal.light.MLight;
	import com.mousman.as3.agal.mesh.IMMesh;
	import com.mousman.as3.agal.MWorld3D;
	import flash.display.Sprite;
	import flash.display.Stage;
	import flash.display3D.Context3D;
	import flash.display3D.Context3DProgramType;
	import flash.display3D.Context3DTriangleFace;
	import flash.display3D.Context3DVertexBufferFormat;
	import flash.display3D.IndexBuffer3D;
	import flash.display3D.VertexBuffer3D;
	import flash.events.Event;
	import flash.events.KeyboardEvent;
	import flash.events.MouseEvent;
	import flash.geom.Matrix3D;
	import flash.geom.Vector3D;
	import flash.text.TextField;
	import flash.text.TextFieldAutoSize;
	import flash.text.TextFormat;
	import flash.utils.getTimer;
	/**
	 * ...
	 * @author Moussa Dembélé (www.mousman.com)
	 */
	public class World extends Sprite
	{
		private var _world3d:MWorld3D;
		private var _cameras:Vector.<IMCamera>;
		private var _meshes:Vector.<IMMesh>;
		private var _projectionMatrix:PerspectiveMatrix3D;
		private var _modelViewProjection:Matrix3D;
		
		private var _ambiantLight:Vector.<Number>;
		
		private var _stage:Stage;
		private var _width:int;
		private var _height:int;
		
		private var _context:Context3D;
		
		private var t:Number = 0;
		
		private var fpsLast:uint = getTimer();
		private var fpsTicks:uint = 0;
		private var fpsTf:TextField;	
		private var looptemp:int = 0;		
		private var _mouseDown:Boolean = false;
		public function World(stage:Stage) 
		{
			_world3d = new MWorld3D();
			// wait for Stage3D to provide us a Context3D
			_stage = stage;
			_width = _stage.stageWidth;
			_height = _stage.stageHeight;
			stage.stage3Ds[0].addEventListener(Event.CONTEXT3D_CREATE, onCreate);
			stage.stage3Ds[0].requestContext3D();			
			
			initGUI();
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function initGUI():void 
		{
			var myFormat:TextFormat = new TextFormat();
			myFormat.color = 0xffffff;
			myFormat.size = 13;
			// create an FPSCounter that displays the framerate on screen
			fpsTf = new TextField();
			fpsTf.x = 0;
			fpsTf.y = 0;
			fpsTf.selectable = false;
			fpsTf.autoSize = TextFieldAutoSize.LEFT;
			fpsTf.defaultTextFormat = myFormat;
			fpsTf.text = "Initializing Stage3d...";
			addChild(fpsTf);			
		}		
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function onCreate(e:Event):void 
		{
			removeEventListener(Event.ENTER_FRAME,render);
			_context = _stage.stage3Ds[0].context3D;
			
			if ( _context == null) return;
			
			_context.configureBackBuffer(_width, _height, 4, true);
			_context.setCulling(Context3DTriangleFace.FRONT);
			// debug mode
			_context.enableErrorChecking = true;
			build();
			setCameraListener();
			addEventListener(Event.ENTER_FRAME, render);		
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function build():void {
			_meshes = new Vector.<IMMesh>();
			_modelViewProjection = new Matrix3D();
			setProjection();
			setMeshs();
			setCamera();
			setLights();
			
			_meshes[0].setSaticConstants();
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function setProjection():void 
		{
			_projectionMatrix = new PerspectiveMatrix3D();
			var aspect:Number = _width /_height; 
			var zNear:Number = 0.1; 
			var zFar:Number = 10000; 
			var fov:Number = 45 * Math.PI / 180; 
			_projectionMatrix.perspectiveFieldOfViewLH(fov, aspect, zNear, zFar);			
		}			
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function setMeshs():void 
		{
			var mesh:CupsMesh = new CupsMesh(_context);
			mesh.position = new Vector3D(0, 0, 100);
			mesh.setRotation(0, 0, 0);
			_meshes.push(mesh);
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function setCamera():void {
			_cameras = new Vector.<IMCamera>;
			var camera:OrbitalCamera = new OrbitalCamera(_meshes[0].modelMatrix,_stage,1);
			_cameras.push(camera);
			
			_meshes[0].cameraMatrix = camera.matrix;
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function setLights():void 
		{
			var light:MLight = new MLight(0xffffff,0.5,0xa2d3e3,0.7);
			light.position = new Vector3D(-50, 120,-50);
			_meshes[0].addLight(light);

			_ambiantLight =  Vector.<Number>([0.2, 0.2, 0.2, 0.1]);			
			_meshes[0].ambiantLight  = _ambiantLight;
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function setCameraListener():void 
		{
			// stage listeners for camera control
			_stage.addEventListener(MouseEvent.MOUSE_DOWN, mouseMove);
			_stage.addEventListener(MouseEvent.MOUSE_UP, mouseMove);
			_stage.addEventListener(MouseEvent.MOUSE_MOVE, mouseMove);
			_stage.addEventListener(MouseEvent.MOUSE_WHEEL, mouseMove);
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function mouseMove(evt:MouseEvent ):void 
		{
			var camera:OrbitalCamera = _cameras[0] as OrbitalCamera;
			switch(evt.type) {
				case MouseEvent.MOUSE_DOWN : 
					camera.mouseDown(); 
					_mouseDown = true;
				break;
				case MouseEvent.MOUSE_UP : 
					camera.mouseUp();  
					_mouseDown = false;
				break;
				case MouseEvent.MOUSE_MOVE : camera.mouseMove(stage.mouseX, stage.mouseY);  break;
				case MouseEvent.MOUSE_WHEEL : camera.mouseWheel(evt.delta);  break;
				case MouseEvent.MOUSE_UP : camera.mouseUp();  break;
				default:break;
			}
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		private function computeMVMatrix(mesh:IMMesh):void 
		{
			// clear the matrix and append new angles
			_modelViewProjection.identity();
			_modelViewProjection.append(mesh.modelMatrix);

			var viewMatrix:Matrix3D = _cameras[0].matrix.clone();
			viewMatrix.invert();
			_modelViewProjection.append(viewMatrix);
			_modelViewProjection.append(_projectionMatrix);				
		}
		//----------------------------------------------------------------------------------------------		
		//----------------------------------------------------------------------------------------------		
		public function render(e:Event ):void 
		{
			if ( !_context ) return;
			_context.clear(0, 0, 0, 1);

			var camera:OrbitalCamera = _cameras[0] as OrbitalCamera;
			camera.update();
			
			// move object
			var mesh:IMMesh = _meshes[0];
			if(!_mouseDown) t += 2.0;			
			var rx:Number = (t * 0.35) % 360 ;
			var ry:Number = (t * 0.3) %360;
			var rz:Number = (t * 0.5) % 360;		
			
			(mesh as CupsMesh).setRotation(rx, ry, rz);
			
			computeMVMatrix(mesh);
			mesh.cameraPosition = camera.position;
			mesh.modelViewProjection = _modelViewProjection;
			// draw current mesh
			mesh.drawTriangles();
			
			_context.present();
			
			// stats
			fpsTicks++;
			var now:uint = getTimer();
			var delta:uint = now - fpsLast;
			// only update the display once a second
			if (delta >= 1000){
				var fps:Number = fpsTicks / delta * 1000;
				fpsTf.text = fps.toFixed(1) + " fps";
				fpsTicks = 0;
				fpsLast = now;
				}				
		}
		
	}

}